@@ -2,7 +2,7 @@
 
 # This module is responsible for the per-connection Broker object
 
-import types
+import types, time
 from itertools import count
 
 from zope.interface import implements
@@ -71,7 +71,7 @@ class PBRootUnslicer(RootUnslicer):
                               ord(typebyte))
             raise BananaError("index token 0x%02x not STRING or VOCAB" % \
                               ord(typebyte))
-        
+
     def open(self, opentype):
         # used for lower-level objects, delegated up from childunslicer.open
         child = RootUnslicer.open(self, opentype)
@@ -157,6 +157,10 @@ class Broker(banana.Banana, referenceable.Referenceable):
             table = vocab.INITIAL_VOCAB_TABLES[vocab_table_index]
             self.populateVocabTable(table)
         self.initBroker()
+        self.current_slave_IR = params.get('current-slave-IR')
+        self.current_seqnum = params.get('current-seqnum')
+        self.current_attempt_id = params.get('current-attempt-id')
+        self.creation_timestamp = time.time()
 
     def initBroker(self):
 
@@ -211,13 +215,28 @@ class Broker(banana.Banana, referenceable.Referenceable):
     # the lack of connection activity
 
     def connectionTimedOut(self):
-        self.shutdown()
+        err = error.ConnectionLost("banana timeout: connection dropped")
+        self.shutdown(err)
 
-    def shutdown(self):
-        self.disconnectWatchers = []
+    def shutdown(self, why, fireDisconnectWatchers=True):
+        """Stop using this connection. If fireDisconnectWatchers is False,
+        all disconnect watchers are removed before shutdown, so they will not
+        be called (this is appropriate when the Broker is shutting down
+        because the whole Tub is being shut down). We terminate the
+        connection quickly, rather than waiting for the transmit queue to
+        drain.
+        """
+        if not fireDisconnectWatchers:
+            self.disconnectWatchers = []
+        self.finish(why)
         self.transport.loseConnection()
 
     def connectionLost(self, why):
+        self.finish(why)
+
+    def finish(self, why):
+        if self.disconnected:
+            return
         self.disconnected = True
         self.remote_broker = None
         self.abandonAllRequests(why)
@@ -300,7 +319,7 @@ class Broker(banana.Banana, referenceable.Referenceable):
         they will give us an interface name too, and possibly a global URL
         for it. Obtain a RemoteReference object (creating it if necessary) to
         give to the local recipient.
-        
+
         The sender remembers that we hold a reference to their object. When
         our RemoteReference goes away, we send a decref message to them, so
         they can possibly free their object. """
@@ -320,7 +339,7 @@ class Broker(banana.Banana, referenceable.Referenceable):
             if url:
                 self.yourReferenceByURL[url] = tracker
         return tracker
-        
+
     def freeYourReference(self, tracker, count):
         # this is called when the RemoteReference is deleted
         if not self.remote_broker: # tests do not set this up
@@ -460,8 +479,7 @@ class Broker(banana.Banana, referenceable.Referenceable):
 
     def abandonAllRequests(self, why):
         for req in self.waitingForAnswers.values():
-            req.fail(why)
-        self.waitingForAnswers = {}
+            eventually(req.fail, why)
 
     # target-side, invoked by CallUnslicer
 
